# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports
import os
import uuid
from pathlib import Path
import logging

# Third Party Imports
import tifffile
import numpy.typing as npt
from numpy import stack

# Local imports
from .data_source import DataSource, DataReader
from ..metadata_sources.metadata import Metadata
from ..metadata_sources.ome_tiff_metadata import OMETIFFMetadata


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class TiffDataSource(DataSource):
    """Data source for TIFF files."""

    def __init__(
        self,
        file_name: str = "",
        mode: str = "w",
        is_bigtiff: bool = False,
    ) -> None:
        """Initialize a TIFF data source.

        Parameters
        ----------
        file_name : str
            Path to file.
        mode : str
            File mode. "r" for read, "w" for write.
        is_bigtiff : bool
            Is this a bigtiff file?
        """
        #: np.ndarray: Image data
        self.image = None
        self._views = []

        super().__init__(file_name=file_name, mode=mode)

        #: str: Directory to save the data to.
        self.save_directory = Path(self.file_name).parent

        # Is this an OME-TIFF?
        # TODO: check the header, rather than use the file extension
        if self.file_name.endswith(".ome.tiff") or self.file_name.endswith(".ome.tif"):
            self._is_ome = True
            #: Metadata: Metadata object
            self.metadata = OMETIFFMetadata()
        else:
            self._is_ome = False
            self.metadata = Metadata()

        self._is_bigtiff = is_bigtiff

        # For file writing, do we assume all files end with tiff or tif?
        self.__double_f = self.file_name.endswith("tiff")

        # Keep track of z, time, channel indices
        # self._current_frame = 0
        self._current_time = 0
        self._current_position = 0

    @property
    def data(self) -> npt.ArrayLike:
        """Return the image data as a numpy array.

        Returns
        -------
        npt.ArrayLike
            Image data.
        """
        self.mode = "r"
        return self.image.asarray()

    @property
    def is_bigtiff(self) -> bool:
        """Is this a bigtiff file?

        Returns
        -------
        bool
            Is this a bigtiff file?
        """
        if self._write_mode:
            return self._is_bigtiff
        else:
            return self.image.is_bigtiff

    def set_bigtiff(self, is_bigtiff: bool) -> None:
        """Set whether this is a bigtiff file.

        Parameters
        ----------
        is_bigtiff : bool
            Is this a bigtiff file?
        """
        self._is_bigtiff = is_bigtiff

    @property
    def is_ome(self) -> bool:
        """Is this an OME-TIFF file?

        Returns
        -------
        bool
            Is this an OME-TIFF file?
        """
        if self._write_mode:
            return self._is_ome
        else:
            return self.image.is_ome

    def read(self) -> None:
        """Read a tiff file."""
        self.mode = "r"
        self.image = tifffile.TiffFile(self.file_name)

        # TODO: Parse metadata
        for i, ax in enumerate(list(self.image.series[0].axes)):
            if ax == "Q":
                # TODO: This is a hack for tifffile. Find a way to remove this.
                ax = "Z"
            setattr(self, f"shape_{ax.lower()}", self.data.shape[i])

    def get_data(
        self,
        timepoint: int = 0,
        position: int = 0,
        channel: int = 0,
        z: int = -1,
        resolution: int = 1,
    ) -> npt.ArrayLike:
        """Get data according to timepoint, position, channel and z-axis id

        Parameters
        ----------
        timepoint : int
            The timepoint value
        position : int
            The position id in multi-position table
        channel : int
            The channel id
        z : int
            The index of Z in a Z-stack.
            Return all z if -1.
        resolution : int
            values from 1, 2, 4, 8
            Not supported for now.

        Returns
        -------
        data : npt.ArrayLike
            Image data
        """
        # TODO: may need to support .tif
        file_suffix = ".ome.tiff" if self.is_ome else ".tiff"
        filename = os.path.join(
            self.save_directory,
            f"Position{position}",
            f"CH{channel:02d}_{timepoint:06d}{file_suffix}",
        )

        if not os.path.exists(filename):
            return None

        image = tifffile.TiffFile(filename)
        if z < 0:
            return TiffReader(image)

        z_num = len(image.pages)
        if z < z_num:
            return image.pages[z].asarray()

        return None

    def write(self, data: npt.ArrayLike, **kw) -> None:
        """Writes 2D image to the data source.

        One channel, all z-position, one timepoint = one stack.
        N channels are opened simultaneously for writing.
        At each time point, a new file is opened for each channel.

        Parameters
        ----------
        data : npt.ArrayLike
            Data to write to file.
        kw : dict
            Keyword arguments to pass to tifffile.imsave.
        """
        self.mode = "w"

        c, z, self._current_time, self._current_position = self._cztp_indices(
            self._current_frame, self.metadata.per_stack
        )  # find current channel
        if z == 0:
            if c == 0:
                # Make sure we're set up for writing
                self._setup_write_image()
            if self.is_ome:
                ome_xml = self.metadata.to_xml(
                    c=c, t=self._current_time, file_name=self.file_name, uid=self.uid
                ).encode()
        else:
            ome_xml = None

        if len(kw) > 0:
            self._views.append(kw)

        if self.is_ome:
            self.image[c].write(data, description=ome_xml, contiguous=True)
        else:
            dx, dy, dz = self.metadata.voxel_size
            md = {
                "spacing": dz,
                "unit": "um",
                "axes": "ZYX",
                "channel": c,
                "timepoint": self._current_time,
                "notes": self.metadata.misc,
            }
            self.image[c].write(
                data,
                resolution=(1e4 / dx, 1e4 / dy, "CENTIMETER"),
                metadata=md,
                contiguous=True,
            )

        self._current_frame += 1

        # Check if this was the last frame to write
        # print("Switch")
        c, z, _, _ = self._cztp_indices(self._current_frame, self.metadata.per_stack)
        if (z == 0) and (c == 0):
            self.close(True)

    def generate_image_name(self, current_channel, current_time_point):
        """Generates a string for the filename, e.g., CH00_000000.tif

        Parameters
        ----------
        current_channel : int
            Current channel index.
        current_time_point : int
            Current time point index.

        Returns
        -------
        str
            Image name.
        """
        ext = ".ome" if self.is_ome else ""
        ext += ".tiff" if self.__double_f else ".tif"
        image_name = (
            "CH0" + str(current_channel) + "_" + str(current_time_point).zfill(6) + ext
        )
        return image_name

    def _setup_write_image(self) -> None:
        """Setup N channel images for writing."""

        # Grab expected data shape from metadata
        (
            self.shape_x,
            self.shape_y,
            self.shape_c,
            self.shape_z,
            self.shape_t,
        ) = self.metadata.shape
        self.dx, self.dy, self.dz = self.metadata.voxel_size
        self.dc, self.dt = self.metadata.dc, self.metadata.dt

        # Initialize one TIFF per channel per time point
        self.image = []
        self.file_name = []
        self.uid = []
        self._views = []

        if self.metadata._multiposition:
            position_directory = os.path.join(
                self.save_directory, f"Position{self._current_position}"
            )
        else:
            position_directory = self.save_directory
        if not os.path.exists(position_directory):
            os.mkdir(position_directory)
        for ch in range(self.shape_c):
            file_name = os.path.join(
                position_directory, self.generate_image_name(ch, self._current_time)
            )
            self.image.append(
                tifffile.TiffWriter(
                    file_name, bigtiff=self.is_bigtiff, ome=False, byteorder="<"
                )
            )
            self.file_name.append(file_name)
            self.uid.append(str(uuid.uuid4()))

    def close(self, internal=False) -> None:
        """Close the file.

        Parameters
        ----------
        internal : bool
            Internal flag. Do not close if True.
        """
        if self._closed and not internal:
            return
        if self.image is None:
            return
        # internal flag needed to avoid _check_shape call until last file is written
        if self.mode == "w":
            if not internal:
                self._check_shape(self._current_frame - 1, self.metadata.per_stack)
        if type(self.image) is list:
            for ch in range(len(self.image)):
                self.image[ch].close()
                if self.is_ome and len(self._views) > 0:
                    # Attach OME metadata at the end of the write
                    tifffile.tiffcomment(
                        self.file_name[ch],
                        self.metadata.to_xml(
                            c=ch,
                            t=self._current_time,
                            file_name=self.file_name,
                            uid=self.uid,
                            views=self._views,
                        ).encode(),
                    )
        else:
            self.image.close()
        if not internal:
            self._closed = True


class TiffReader(DataReader):
    def __init__(self, tiff_file: tifffile.TiffFile):
        self.tiff = tiff_file

    @property
    def shape(self):
        page_number = len(self.tiff.pages)

        x, y = self.tiff.pages[0].shape

        return (page_number, x, y)

    def __getitem__(self, index):

        if isinstance(index, int):
            # Return the entire page as a NumPy array
            return self.tiff.pages[index].asarray()

        elif isinstance(index, slice):
            # Handle case for slicing all pages
            pages = [
                self.tiff.pages[i].asarray() for i in range(index.start, index.stop)
            ]
            return stack(pages, axis=0)

        elif isinstance(index, tuple):
            # Check if the first index is an integer (page index)
            if isinstance(index[0], int):
                page_index = index[0]
                if len(index) == 2:
                    return self.tiff.pages[page_index].asarray()[index[1]]
                elif len(index) == 3:
                    return self.tiff.pages[page_index].asarray()[index[1], index[2]]
            elif isinstance(index[0], slice):
                if len(index) == 2:
                    pages = [
                        self.tiff.pages[i].asarray()[index[1]]
                        for i in range(index.start, index.stop)
                    ]

                elif len(index) == 3:
                    pages = [
                        self.tiff.pages[i].asarray()[index[1], index[2]]
                        for i in range(index.start, index.stop)
                    ]
                return stack(pages, axis=0)

        logger.debug(f"TiffReader: Invalid indexing format. {index}")
        return None

    def __array__(self):
        return self.tiff.asarray()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
from typing import Any, Dict


# Local Imports
from navigate.model.devices.galvo.base import GalvoBase
from navigate.model.devices.device_types import SerialDevice
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController
from navigate.tools.decorators import log_initialization

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class ASIGalvo(GalvoBase, SerialDevice):
    """GalvoASI Class - ASI DAQ Control of Galvanometers"""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        device_id: int = 0,
    ) -> None:
        """Initialize the GalvoASI class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            Connection to the NI DAQ device.
        configuration : Dict[str, Any]
            Dictionary of configuration parameters.
        device_id : int
            Galvo ID. Default is 0.
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        #: Any: Device connection.
        self.galvo = device_connection

        #: str: Galvo Axis
        self.axis = self.device_config["hardware"]["axis"]  # .get("axis","B")
        logger.debug(f"galvo axis: {self.axis}")

    def __str__(self) -> str:
        """Return string representation of the GalvoASI."""
        return "GalvoASI"

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASILaser Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, default is 115200.
        timeout : float
            Timeout for communicating with the filter wheel, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI stage connection failed.")
            raise Exception("ASI stage connection failed.")
        return tiger_controller

    def adjust(self, exposure_times, sweep_times):
        """Adjust the galvo waveform to account for the camera readout time.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of camera exposure time in seconds on a per-channel basis.
            e.g., exposure_times = {"channel_1": 0.1, "channel_2": 0.2}
        sweep_times : dict
            Dictionary of acquisition sweep time in seconds on a per-channel basis.
            e.g., sweep_times = {"channel_1": 0.1, "channel_2": 0.2}

        Returns
        -------
        waveform_dict : dict
            Dictionary that includes the galvo waveforms on a per-channel basis.
        """

        # TODO: variables galvo_amplitude, offset, etc., are local in the parent
        #  method and not assigned as an attribute. As such, much of this code is
        #  duplicated. We could consider refactoring this to avoid duplication.

        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        microscope_name = microscope_state["microscope_name"]
        zoom_value = microscope_state["zoom"]
        galvo_factor = self.configuration["waveform_constants"]["other_constants"].get(
            "galvo_factor", "none"
        )
        galvo_parameters = self.configuration["waveform_constants"]["galvo_constants"][
            self.galvo_name
        ][microscope_name][zoom_value]
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["sample_rate"]

        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:

                # Get the Waveform Parameters - Assumes ETL Delay < Camera Delay.
                # Should Assert.
                exposure_time = exposure_times[channel_key]
                self.sweep_time = sweep_times[channel_key]

                # galvo Parameters
                try:
                    galvo_amplitude = float(galvo_parameters.get("amplitude", 0))
                    galvo_offset = float(galvo_parameters.get("offset", 0))
                    galvo_rising_ramp = float(galvo_parameters.get("rising_ramp", 50))
                    if float(galvo_parameters.get("frequency", 10)) == 0:
                        galvo_period = exposure_time
                    else:
                        galvo_period = exposure_time / float(
                            galvo_parameters.get("frequency", 10)
                        )
                    factor_name = None
                    if galvo_factor == "channel":
                        factor_name = (
                            f"Channel {channel_key[channel_key.index('_')+1:]}"
                        )
                    elif galvo_factor == "laser":
                        factor_name = channel["laser"]
                    if factor_name and factor_name in galvo_parameters.keys():
                        galvo_amplitude = float(
                            galvo_parameters[factor_name].get("amplitude", 0)
                        )
                        galvo_offset = float(
                            galvo_parameters[factor_name].get("offset", 0)
                        )

                except ValueError as e:
                    logger.debug(
                        f"{e} waveform constants.yml doesn't have parameter "
                        f"amplitude/offset/frequency for {self.galvo_name}"
                    )
                    return

                # Calculate the Waveforms
                if self.galvo_waveform == "sawtooth":
                    period = galvo_period
                    amplitude = galvo_amplitude
                    offset = galvo_offset
                    duty_cycle = galvo_rising_ramp

                    # Duty cycle must be either 0, 50, or 100
                    # If the duty cycle is not 0, 50, or 100, it will round to the nearest value
                    if duty_cycle not in (0, 50, 100):
                        duty_cycle = round(duty_cycle / 50) * 50
                        logger.debug(
                            f"Invalid duty cycle given. Duty cycle value corrected to {duty_cycle}"
                        )
                    self.configuration["waveform_constants"]["galvo_constants"][
                        self.galvo_name
                    ][microscope_name][zoom_value]["rising ramp"] = duty_cycle
                    print(
                        self.configuration["waveform_constants"]["galvo_constants"][
                            self.galvo_name
                        ][microscope_name][zoom_value]["rising ramp"]
                    )
                    self.sawtooth(period, amplitude, offset, duty_cycle)

                elif self.galvo_waveform == "sine":
                    period = galvo_period
                    amplitude = galvo_amplitude
                    offset = galvo_offset

                    self.sine_wave(period, amplitude, offset)

                else:
                    print("Unknown Galvo waveform specified in configuration file.")
                    continue

    def sawtooth(self, period=10, amplitude=1, offset=0, duty_cycle=100):
        """
        Sends the tiger controller commands to initiate the sawtooth wave.

        If the duty cycle given is 50, a triangle wave will be initiated.

        Parameters
        ----------
        period : Float
            Unit - milliseconds
        amplitude : Float
            Unit - Volts
        offset : Float
            Unit - Volts
        duty_cycle : Float
            Unit - Percent
        """

        # Converts period to ms and amplitude and offset to mV
        period = int(round(period * 1000))
        amplitude *= 1000
        offset *= 1000

        if duty_cycle == 0:
            # Negative amplitude reverses the polarity of the waveform
            amplitude = amplitude * -1
            # Sawtooth waveform that is triggered on TTL input
            self.galvo.SA_waveform(self.axis, 128, amplitude, offset, period)

        if duty_cycle == 50:
            # Adjusts the period for a triangle waveform
            period = 2 * round(period / 2)
            # Triangle waveform that is triggered on TTL input
            self.galvo.SA_waveform(self.axis, 129, amplitude, offset, period)

        if duty_cycle == 100:
            # Sawtooth waveform that is triggered on TTL input
            self.galvo.SA_waveform(self.axis, 128, amplitude, offset, period)

        # Waveform is free running after TTL input
        self.galvo.SAM(self.axis, 4)

    def sine_wave(self, period=10.0, amplitude=1.0, offset=0.0):
        """Sends the tiger controller commands to initiate the sine wave.

        Parameters
        ----------
        period : Float
            Unit - milliseconds
        amplitude : Float
            Unit - Volts
        offset : Float
            Unit - Volts

        """

        # Converts period to ms and amplitude and offset to mV
        period = int(round(period * 1000))
        amplitude *= 1000
        offset *= 1000

        # Sine wave that is triggered on TTL input
        self.galvo.SA_waveform(self.axis, 131, amplitude, offset, period)
        # Waveform is free running after it is triggered
        self.galvo.SAM(self.axis, 4)

    def turn_off(self):
        """Stops the galvo waveform"""
        self.galvo.SAM(self.axis, 0)

    def close(self):
        """Close the ASI galvo serial port.

        Stops the remote focus waveform and then closes the port.
        """
        if self.galvo.is_open():
            self.turn_off()
            logger.debug("ASI Remote Focus - Closing Device.")
            self.galvo.disconnect_from_serial()

    def __del__(self):
        """Destructor for the ASIGalvo class."""
        self.close()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import logging
from typing import Any, Dict

# Third-party imports
import numpy as np
import numpy.typing as npt

# Local application imports
from .data_source import DataSource
from ...tools.slicing import ensure_slice, ensure_iter, slice_len

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class PyramidalDataSource(DataSource):
    """General class for data sources that store data in a pyramidal structure.

    Implements resolution/subdivision calculations and __getitem__ with
    indexing by subdivision.
    """

    def __init__(self, file_name: str = None, mode: str = "w") -> None:
        """Initializes the PyramidalDataSource.

        Parameters
        ----------
        file_name : str
            The name of the file to write to.
        mode : str
            The mode to open the file in. Must be "w" for write or "r" for read.
        """
        #: np.array: The resolution of each down-sampled pyramid level.
        self._resolutions = np.array([[1, 1, 1]], dtype=int)
        #: np.array: The number of subdivisions in each dimension.
        self._subdivisions = None

        #: np.array: The shape of the image.
        self._shapes = None

        super().__init__(file_name, mode)

    @property
    def resolutions(self) -> npt.NDArray:
        """Getter for resolutions.

        Store as XYZ per BDV spec.

        Returns
        -------
        resolutions : npt.NDArray
            The resolutions.
        """
        return self._resolutions

    @property
    def subdivisions(self) -> npt.NDArray:
        """Getter for subdivisions.

        Store as XYZ per BDV spec.

        Returns
        -------
        subdivisions : npt.NDArray
            The subdivisions.
        """
        if self._subdivisions is None:
            self._subdivisions = np.zeros((self.resolutions.shape[0], 3), dtype=int)
            self._subdivisions[:, 0] = np.gcd(32, self.shapes[:, 0])
            self._subdivisions[:, 1] = np.gcd(32, self.shapes[:, 1])
            self._subdivisions[:, 2] = np.gcd(32, self.shapes[:, 2])

            # Safety
            self._subdivisions = np.maximum(self._subdivisions, 1)

            # Reverse to XYZ
            self._subdivisions = self._subdivisions[:, ::-1]
        return self._subdivisions

    @property
    def shapes(self) -> npt.NDArray:
        """Getter for image shape.

        Store as ZYX rather than XYZ, per BDV spec.

        Returns
        -------
        shapes : npt.NDArray
            The shapes.
        """
        if self._shapes is None:
            self._shapes = np.maximum(
                np.ceil(
                    np.array([self.shape_z, self.shape_y, self.shape_x])[None, :]
                    / self.resolutions[:, ::-1]
                ).astype(int),
                1,
            )
        return self._shapes

    @property
    def nbytes(self) -> int:
        """Getter for image size.

        Size in bytes. Overrides base class. Accounts for subdivisions.

        Returns
        -------
        size : int
            The size of the image in bytes.
        """
        return (
            np.prod(self.shapes, axis=1)
            * self.shape_t
            * self.shape_c
            * self.positions
            * self.bits
            // 8
        ).sum()

    def set_metadata_from_configuration_experiment(
        self, configuration: Dict[str, Any], microscope_name: str = None
    ) -> None:
        """Sets the metadata from according to the microscope configuration.

        Parameters
        ----------
        configuration : Dict[str, Any]
            The configuration experiment.
        microscope_name : str
            The microscope name
        """
        self._subdivisions = None
        self._shapes = None

        if ("BDVParameters" in configuration["experiment"].keys()
            and "down_sample" in configuration["experiment"]["BDVParameters"].keys()):
            down_sample = configuration["experiment"]["BDVParameters"]["down_sample"].get(
                "down_sample", False
            )

            if down_sample:
                max_xy = configuration["experiment"]["BDVParameters"]["down_sample"].get(
                    "lateral_down_sample", 1
                )
                max_z = configuration["experiment"]["BDVParameters"]["down_sample"].get(
                    "axial_down_sample", 1
                )

                xy_values = [2**i for i in range(int(np.log2(max_xy)) + 1)]
                z_values = [2**i for i in range(int(np.log2(max_z)) + 1)]

                max_len = max(len(xy_values), len(z_values))
                xy_values.extend([xy_values[-1]] * (max_len - len(xy_values)))
                z_values.extend([z_values[-1]] * (max_len - len(z_values)))

                #: npt.NDArray: The resolution of each down-sampled pyramid level.
                self._resolutions = np.array(
                    [[xy, xy, z] for xy, z in zip(xy_values, z_values)], dtype=int
                )

        return super().set_metadata_from_configuration_experiment(
            configuration, microscope_name
        )

    def __getitem__(self, keys):
        """Magic method to get slice requests passed by, e.g., ds[:,2:3,...].
        Allows arbitrary slicing of dataset via calls to get_slice().

        Order is xycztps where x, y, z are array indices, c is channel,
        t is timepoints, p is positions and s is subdivisions to index along.

        TODO: Add subdivisions.

        Parameters
        ----------
        keys : tuple
            Tuple of indices.

        Returns
        -------
        npt.ArrayLike
            Array of shape (p, t, z, c, y, x)
        """

        # Check lengths
        if isinstance(keys, slice) or isinstance(keys, int):
            length = 1
        else:
            length = len(keys)

        if length < 1:
            error_statement = (
                "Too few indices. Indices may be (x, y, c, z, t, p, subdivisions)."
            )
            logger.error(error_statement)
            raise IndexError(error_statement)
        elif length > 7:
            error_statement = (
                "Too many indices. Indices may be (x, y, c, z, t, p, subdivisions)."
            )
            logger.error(error_statement)
            raise IndexError(error_statement)

        # Get indices as slices/ranges
        xs = ensure_slice(keys, 0)
        ys = ensure_slice(keys, 1)
        cs = ensure_iter(keys, 2, self.shape[2])
        zs = ensure_slice(keys, 3)
        ts = ensure_iter(keys, 4, self.shape[4])
        ps = ensure_iter(keys, 5, self.positions)

        if length > 1 and keys[-1] == Ellipsis:
            keys = keys[:-1]
            length -= 1

        if length > 6 and isinstance(keys[6], int):
            sub_divisions = keys[6]
        else:
            sub_divisions = 0

        if len(cs) == 1 and len(ts) == 1 and len(ps) == 1:
            return self.get_slice(xs, ys, cs[0], zs, ts[0], ps[0], sub_divisions)

        sliced_ds = np.empty(
            (
                len(ps),
                len(ts),
                slice_len(zs, self.shape_z) // self.resolutions[sub_divisions][2],
                len(cs),
                slice_len(ys, self.shape_y) // self.resolutions[sub_divisions][1],
                slice_len(xs, self.shape_x) // self.resolutions[sub_divisions][0],
            ),
            dtype=self.dtype,
        )

        for c in cs:
            for t in ts:
                for p in ps:
                    sliced_ds[p, t, :, c, :, :] = self.get_slice(
                        xs, ys, c, zs, t, p, sub_divisions
                    )

        return sliced_ds

    def get_slice(self, x, y, c, z=0, t=0, p=0, subdiv=0) -> npt.ArrayLike:
        """Get a 3D slice of the dataset for a single c, t, p, subdiv.

        Parameters
        ----------
        x : int or slice
            x indices to grab
        y : int or slice
            y indices to grab
        c : int
            Single channel
        z : int or slice
            z indices to grab
        t : int
            Single timepoint
        p : int
            Single position
        subdiv : int
            Subdivision of the dataset to index along

        Returns
        -------
        npt.ArrayLike
            3D (z, y, x) slice of data set

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a derived class.
        """
        error_statement = "Implemented in a derived class."
        logger.error(error_statement)
        raise NotImplementedError(error_statement)



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Imports
import threading
import time
import logging

# Third Party Imports
from serial import Serial
from serial import SerialException
from serial import SerialTimeoutException
from serial import EIGHTBITS
from serial import PARITY_NONE
from serial import STOPBITS_ONE

# Local Imports
from navigate.model.devices.APIs.asi.asi_tiger_controller import ASIException, TigerController
# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MS2000Controller(TigerController):
    """MS2000 Controller class"""

    def __init__(self, com_port: str, baud_rate: int, verbose: bool = False):
        """Initialize the MS2000 Controller class


        Parameters
        ----------
        com_port : str
            COM port of the MS2000 Controller
        baud_rate : int
            Baud rate of the MS2000 Controller
        verbose : bool
            If True, will print out messages to the console

        """
        super().__init__(com_port, baud_rate, verbose)

    def __str__(self) -> str:
        """Returns the string representation of the MS2000 Controller class"""
        return "MS2000Controller"


    def connect_to_serial(
        self,
        rx_size: int = 12800,
        tx_size: int = 12800,
        read_timeout: int = 1,
        write_timeout: int = 1,
    ) -> None:
        """
        Connect to the serial port.

        Parameters
        ----------
        rx_size : int
            Size of the rx buffer
        tx_size : int
            Size of the tx buffer
        read_timeout : int
            Read timeout in seconds
        write_timeout : int
            Write timeout in seconds
        """
        self.serial_port.port = self.com_port
        self.serial_port.baudrate = self.baud_rate
        self.serial_port.parity = PARITY_NONE
        self.serial_port.bytesize = EIGHTBITS
        self.serial_port.stopbits = STOPBITS_ONE
        self.serial_port.xonoff = False
        self.serial_port.rtscts = False
        self.serial_port.dsrdtr = False
        self.serial_port.write_timeout = write_timeout
        self.serial_port.timeout = read_timeout

        # set the size of the rx and tx buffers before calling open
        self.serial_port.set_buffer_size(rx_size, tx_size)
        try:
            self.serial_port.open()
        except SerialException:
            self.report_to_console(
                f"SerialException: can't connect to {self.com_port} at "
                f"{self.baud_rate}!"
            )

        if self.is_open():
            # clear the rx and tx buffers
            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()
            # report connection status to user
            self.report_to_console("Connected to the serial port.")
            self.report_to_console(
                f"Serial port = {self.com_port} :: Baud rate = {self.baud_rate}"
            )

            #: list[str]: Default axes sequence of the MS2000 Controller
            self.default_axes_sequence = [
                "X",
                "Y",
                "Z",
            ]  # self.get_default_motor_axis_sequence()

    def get_default_motor_axis_sequence(self) -> None:
        """Get the default motor axis sequence from the ASI device

        Returns
        -------
        list[str]
            Default motor axis sequence
        """
        self.send_command("BU X")
        response = self.read_response()
        lines = response.split("\r")
        for line in lines:
            if line.startswith("Motor Axes:"):
                default_axes_sequence = line[line.index(":") + 2 :].split(" ")[:-2]
                self.report_to_console(
                    "Get the default axes sequence from the ASI device " "successfully!"
                )
                break

        return default_axes_sequence


    def wait_for_device(self, report: bool = False):
        """Waits for the all motors to stop moving."""
        if not report:
            print("Waiting for device...")
        temp = self.report
        self.report = report
        busy = True
        while busy:
            busy = self.is_device_busy()
        self.report = temp

    def send_command(self, cmd: str) -> None:
        """Send a serial command to the device.

        Parameters
        ----------
        cmd : str
            Serial command to send to the device
        """
        # always reset the buffers before a new command is sent
        self.safe_to_write.wait()
        self.safe_to_write.clear()
        self.serial_port.read_all()
        self.serial_port.reset_input_buffer()
        self.serial_port.reset_output_buffer()

        # send the serial command to the controller
        self.report_to_console(cmd)
        command = bytes(f"{cmd}\r", encoding="ascii")
        try:
            self.serial_port.write(command)
        except SerialTimeoutException as e:
            print(f"MS2000 Controller -- SerialTimeoutException: {e}")
            pass

        # sleep to avoid error, empirically found this made it work
        time.sleep(0.1)


    def set_max_speed(self, axis: str, speed: float) -> None:
        """Set the speed on a specific axis. Speed is in mm/s.

        Parameters
        ----------
        axis : str
            Stage axis
        speed : float
            Speed in mm/s
        """
        self.send_command("/")
        self.read_response()

        self.send_command(f"SPEED {axis}={speed}\r")
        self.read_response()

    # Utility Functions

    def set_triggered_move(self, axis):
        """
        Set ASI to repeat the latest relative movement
        upon receiving a TTL pulse

        Parameters
        ----------
        axis : str
            Stage axis
        """
        self.send_command(f"TTL {axis}=2")
        self.read_response()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


#  Standard Library Imports
from multiprocessing import Manager
import tkinter
from tkinter import messagebox
import multiprocessing as mp
import threading
import sys
import os
import time
import platform

# Third Party Imports

# Local View Imports
from navigate.view.main_application_window import MainApp as view
from navigate.view.popups.camera_view_popup_window import CameraViewPopupWindow
from navigate.view.popups.feature_list_popup import FeatureListPopup

# Local Sub-Controller Imports
from navigate.controller.configuration_controller import ConfigurationController
from navigate.controller.sub_controllers import (
    KeystrokeController,
    WaveformTabController,
    StageController,
    CameraSettingController,
    CameraViewController,
    MIPViewController,
    MultiPositionController,
    ChannelsTabController,
    AcquireBarController,
    FeaturePopupController,
    MenuController,
    PluginsController,
    HistogramController,
    # MicroscopePopupController,
    # AdaptiveOpticsPopupController,
)

from navigate.controller.thread_pool import SynchronizedThreadPool

# Local Model Imports
from navigate.model.model import Model
from navigate.model.concurrency.concurrency_tools import ObjectInSubprocess

# Misc. Local Imports
from navigate._commit import get_git_revision_hash, get_version_from_file
from navigate.config.config import (
    load_configs,
    update_config_dict,
    verify_experiment_config,
    verify_waveform_constants,
    verify_positions_config,
    verify_configuration,
    get_navigate_path,
)
from navigate.tools.file_functions import (
    load_yaml_file,
    save_yaml_file,
    get_ram_info,
)
from navigate.tools.common_dict_tools import update_stage_dict
from navigate.tools.multipos_table_tools import update_table
from navigate.tools.common_functions import combine_funcs

# Logger Setup
import logging

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Controller:
    """Navigate Controller"""

    def __init__(
        self,
        root,
        splash_screen,
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
        gui_configuration_path,
        multi_positions_path,
        args,
    ):
        """Initialize the Navigate Controller.

        Parameters
        ----------
        root : Tk top-level widget.
            Tk.tk GUI instance.
        splash_screen : Tk top-level widget.
            Tk.tk GUI instance.
        configuration_path : string
            Path to the configuration yaml file.
            Provides global microscope configuration parameters.
        experiment_path : string
            Path to the experiment yaml file.
            Provides experiment-specific microscope configuration.
        waveform_constants_path : string
            Path to the waveform constants yaml file.
            Provides magnification and wavelength-specific parameters.
        rest_api_path : string
            Path to the REST API yaml file.
            Provides REST API configuration parameters.
        waveform_templates_path : string
            Path to the waveform templates yaml file.
            Provides waveform templates for each channel.
        gui_configuration_path : string
            Path to the GUI configuration yaml file.
            Provides GUI configuration parameters.
        *args :
            Command line input arguments for non-default
            file paths or using synthetic hardware modes.
        """
        logger.info(f"Navigate GIT Hash: {get_git_revision_hash()}")
        logger.info(f"Navigate Version: {get_version_from_file()}")

        #: Tk top-level widget: Tk.tk GUI instance.
        self.root = root
        #: bool: Flag to indicate if the GUI is ready for resizing.
        self.resize_ready_flag = False

        #: Tk top-level widget: Tk.tk GUI instance.
        self.splash_screen = splash_screen

        #: string: Path to the configuration yaml file.
        self.configuration_path = configuration_path
        logger.info(f"Configuration Path: {self.configuration_path}")

        #: string: Path to the experiment yaml file.
        self.experiment_path = experiment_path
        logger.info(f"Experiment Path: {self.experiment_path}")

        #: string: Path to the waveform constants yaml file.
        self.waveform_constants_path = waveform_constants_path
        logger.info(f"Waveform Constants Path: {self.waveform_constants_path}")

        #: string: Path to the REST API yaml file.
        self.rest_api_path = rest_api_path
        logger.info(f"REST API Path: {self.rest_api_path}")

        #: string: Path to the waveform templates yaml file.
        self.waveform_templates_path = waveform_templates_path
        logger.info(f"Waveform Templates Path: {self.waveform_templates_path}")

        #: string: Path to the GUI configuration yaml file.
        self.gui_configuration_path = gui_configuration_path
        logger.info(f"GUI Configuration Path: {self.gui_configuration_path}")

        #: iterable: Non-default command line input arguments for
        self.args = args
        logger.info(f"Variable Input Arguments: {self.args}")

        #: Object: Thread pool for the controller.
        self.threads_pool = SynchronizedThreadPool()

        #: mp.Queue: Queue for retrieving events ('event_name', value) from model
        self.event_queue = mp.Queue(100)

        #: Manager: A shared memory manager
        self.manager = Manager()

        #: dict: Configuration dictionary
        self.configuration = load_configs(
            self.manager,
            configuration=self.configuration_path,
            experiment=self.experiment_path,
            waveform_constants=self.waveform_constants_path,
            rest_api_config=self.rest_api_path,
            waveform_templates=self.waveform_templates_path,
            gui=self.gui_configuration_path,
        )

        verify_configuration(self.manager, self.configuration)
        verify_experiment_config(self.manager, self.configuration)
        verify_waveform_constants(self.manager, self.configuration)

        positions = load_yaml_file(multi_positions_path)
        positions = verify_positions_config(positions)
        self.configuration["multi_positions"] = positions

        total_ram, available_ram = get_ram_info()
        logger.info(
            f"Total RAM: {total_ram / 1024**3:.2f} GB. "
            f"Available RAM: {available_ram / 1024**3:.2f} GB."
        )

        #: ObjectInSubprocess: Model object in MVC architecture.
        self.model = ObjectInSubprocess(
            Model, args, self.configuration, event_queue=self.event_queue
        )

        #: mp.Pipe: Pipe for sending images from model to view.
        self.show_img_pipe = self.model.create_pipe("show_img_pipe")

        #: string: Path to the default experiment yaml file.
        self.default_experiment_file = self.experiment_path

        #: string: Path to the waveform constants yaml file.
        self.waveform_constants_path = waveform_constants_path

        #: ConfigurationController: Configuration Controller object.
        self.configuration_controller = ConfigurationController(self.configuration)

        #: View: View object in MVC architecture.
        self.view = view(self.root)

        #: dict: Event listeners for the controller.
        self.event_listeners = {}

        #: AcquireBarController: Acquire Bar Sub-Controller.
        self.acquire_bar_controller = AcquireBarController(self.view.acquire_bar, self)

        #: ChannelsTabController: Channels Tab Sub-Controller.
        self.channels_tab_controller = ChannelsTabController(
            self.view.settings.channels_tab, self
        )

        #: MultiPositionController: Multi-Position Tab Sub-Controller.
        self.multiposition_tab_controller = MultiPositionController(
            self.view.settings.multiposition_tab.multipoint_list, self
        )

        #: CameraViewController: Camera View Tab Sub-Controller.
        self.camera_view_controller = CameraViewController(
            self.view.camera_waveform.camera_tab, self
        )

        self.histogram_controller = HistogramController(
            self.view.camera_waveform.camera_tab.histogram, self
        )

        #: MIPSettingController: MIP Settings Tab Sub-Controller.
        self.mip_setting_controller = MIPViewController(
            self.view.camera_waveform.mip_tab, self
        )

        #: CameraSettingController: Camera Settings Tab Sub-Controller.
        self.camera_setting_controller = CameraSettingController(
            self.view.settings.camera_settings_tab, self
        )

        #: StageController: Stage Sub-Controller.
        self.stage_controller = StageController(
            self.view.settings.stage_control_tab,
            self.view,
            self.camera_view_controller.canvas,
            self,
        )

        #: WaveformTabController: Waveform Display Sub-Controller.
        self.waveform_tab_controller = WaveformTabController(
            self.view.camera_waveform.waveform_tab, self
        )

        #: KeystrokeController: Keystroke Sub-Controller.
        self.keystroke_controller = KeystrokeController(self.view, self)

        # Exit the program when the window is closed
        self.view.root.protocol(
            "WM_DELETE_WINDOW", self.acquire_bar_controller.exit_program
        )

        # Bonus config
        self.update_acquire_control()

        t = threading.Thread(target=self.update_event)
        t.start()

        #: MenuController: Menu Sub-Controller.
        self.menu_controller = MenuController(view=self.view, parent_controller=self)
        self.menu_controller.initialize_menus()

        #: dict: acquisition modes from plugins
        self.plugin_acquisition_modes = {}

        #: PluginsController: Plugin Sub-Controller
        self.plugin_controller = PluginsController(
            view=self.view, parent_controller=self
        )
        self.plugin_controller.load_plugins()

        #: int: Number of x_pixels from microscope configuration file.
        self.img_width = 0

        #: int: Number of y_pixels from microscope configuration file.
        self.img_height = 0

        #: SharedNDArray: Pre-allocated shared memory array.
        self.data_buffer = None

        #: dict: Additional microscopes.
        self.additional_microscopes = {}

        #: dict: Additional microscope configurations.
        self.additional_microscopes_configs = {}

        #: bool: Flag for stopping acquisition.
        self.stop_acquisition_flag = False

        #: int: current image id in the buffer
        self.current_image_id = -1

        # Set view based on model.experiment
        self.populate_experiment_setting(in_initialize=True)

        # Camera View Tab
        self.initialize_cam_view()

        # destroy splash screen and show main screen
        self.splash_screen.destroy()
        self.root.deiconify()

        #: int: ID for the resize event.Only works on Windows OS.
        self.resize_event_id = None
        self.window_width = 0
        self.window_height = 0
        self.view.root.after(5000, self.enable_resize)
        self.view.root.bind("<Configure>", self.resize)

    def update_buffer(self):
        """Update the buffer size according to the camera
        dimensions listed in the experimental parameters.

        Returns
        -------
        self.img_width : int
            Number of x_pixels from microscope configuration file.
        self.image_height : int
            Number of y_pixels from microscope configuration file.
        self.data_buffer : SharedNDArray
            Pre-allocated shared memory array.
            Size dictated by x_pixels, y_pixels, an number_of_frames in
            configuration file.
        """
        microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]
        img_width = int(
            self.configuration["experiment"]["CameraParameters"][microscope_name][
                "img_x_pixels"
            ]
        )
        img_height = int(
            self.configuration["experiment"]["CameraParameters"][microscope_name][
                "img_y_pixels"
            ]
        )
        if img_width == self.img_width and img_height == self.img_height:
            return

        if self.data_buffer is not None:
            for i in range(len(self.data_buffer)):
                self.data_buffer[i].shared_memory.close()
        self.data_buffer = self.model.get_data_buffer(img_width, img_height)
        self.img_width = img_width
        self.img_height = img_height

    def update_acquire_control(self):
        """Update the acquire control based on the current experiment parameters."""
        self.view.acquire_bar.stop_stage.config(
            command=self.stage_controller.stop_button_handler
        )

    def change_microscope(self, microscope_name, zoom=None):
        """Change the microscope configuration.

        Parameters
        ----------
        microscope_name : string
            Name of the microscope to change to.
        zoom : string
            Name of the zoom value to change to.
        """
        self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ] = microscope_name
        if zoom:
            self.configuration["experiment"]["MicroscopeState"]["zoom"] = zoom
        if self.configuration_controller.change_microscope():
            # update widgets
            self.stage_controller.initialize()
            self.channels_tab_controller.initialize()
            self.channels_tab_controller.populate_experiment_values()
            self.camera_setting_controller.update_camera_device_related_setting()
            self.camera_setting_controller.populate_experiment_values()
            self.camera_setting_controller.calculate_physical_dimensions()
            self.camera_view_controller.update_snr()

        if (
            hasattr(self, "waveform_popup_controller")
            and self.waveform_popup_controller
        ):
            self.waveform_popup_controller.populate_experiment_values()

    def initialize_cam_view(self):
        """Populate view and maximum intensity projection tabs.

        Communicates with the camera view controller and mip setting controller to
        set the minimum and maximum counts, as well as the default channel settings.
        """
        # Populating Min and Max Counts
        self.camera_view_controller.initialize("minmax", [0, 2**16 - 1])
        self.mip_setting_controller.initialize("minmax", [0, 2**16 - 1])
        self.camera_view_controller.initialize("image", [1, 0, 0])

    def populate_experiment_setting(self, file_name=None, in_initialize=False):
        """Load experiment file and populate model.experiment and configure view.

        Confirms that the experiment file exists.
        Sends the experiment file to the model and the controller.
        Populates the GUI with these settings.

        Parameters
        __________
        file_name : string
            file_name = path to the non-default experiment yaml file.

        """
        # read the new file and update info of the configuration dict
        if not in_initialize:
            update_config_dict(
                self.manager, self.configuration, "experiment", file_name
            )
            verify_experiment_config(self.manager, self.configuration)

        # update buffer
        self.update_buffer()

        # Configure GUI
        microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]
        self.configuration_controller.change_microscope()
        self.menu_controller.resolution_value.set(
            f"{microscope_name} "
            f"{self.configuration['experiment']['MicroscopeState']['zoom']}"
        )
        self.menu_controller.disable_stage_limits.set(
            0 if self.configuration["experiment"]["StageParameters"]["limits"] else 1
        )
        self.execute(
            "stage_limits",
            self.configuration["experiment"]["StageParameters"]["limits"],
        )

        self.acquire_bar_controller.populate_experiment_values()
        # self.stage_controller.populate_experiment_values()
        self.multiposition_tab_controller.set_positions(
            self.configuration["multi_positions"]
        )
        self.channels_tab_controller.populate_experiment_values()
        self.camera_setting_controller.populate_experiment_values()
        self.waveform_tab_controller.set_waveform_template(
            self.configuration["experiment"]["MicroscopeState"]["waveform_template"]
        )

        # autofocus popup
        if hasattr(self, "af_popup_controller"):
            self.af_popup_controller.populate_experiment_values()

        if file_name:
            self.plugin_controller.populate_experiment_setting()

        # set widget modes
        self.set_mode_of_sub("stop")
        self.stage_controller.initialize()

    def update_experiment_setting(self):
        """Update model.experiment according to values in the GUI

        Collect settings from sub-controllers will validate the value, if something
        is wrong, it will return False

        Returns
        -------
        string
            Warning info if any

        """
        microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]
        zoom_value = self.configuration["experiment"]["MicroscopeState"]["zoom"]
        resolution_value = self.menu_controller.resolution_value.get()

        # set microscope and zoom value according to GUI
        if f"{microscope_name} {zoom_value}" != resolution_value:
            microscope_name, zoom_value = resolution_value.split()
            self.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ] = microscope_name
            self.configuration["experiment"]["MicroscopeState"]["zoom"] = zoom_value
            self.execute("resolution", resolution_value)

        warning_message = self.camera_setting_controller.update_experiment_values()

        # set waveform template
        if self.acquire_bar_controller.mode in ["live", "single", "z-stack"]:
            camera_setting = self.configuration["experiment"]["CameraParameters"][
                microscope_name
            ]
            if camera_setting["sensor_mode"] == "Light-Sheet" and camera_setting[
                "readout_direction"
            ] in ["Bidirectional", "Rev. Bidirectional"]:
                self.waveform_tab_controller.set_waveform_template("Bidirectional")
            else:
                self.waveform_tab_controller.set_waveform_template("Default")

        # update multi-positions
        positions = self.multiposition_tab_controller.get_positions()
        self.configuration["multi_positions"] = positions

        if (
            self.configuration["experiment"]["MicroscopeState"]["is_multiposition"]
            and len(positions) == 0
        ):
            # Update the view and override the settings.
            self.configuration["experiment"]["MicroscopeState"][
                "is_multiposition"
            ] = False
            self.channels_tab_controller.is_multiposition_val.set(False)

        # TODO: validate experiment dict
        self.channels_tab_controller.update_experiment_values()
        warning_message += self.channels_tab_controller.verify_experiment_values()

        # additional microscopes
        for microscope_name in self.additional_microscopes_configs:
            if hasattr(self, f"{microscope_name.lower()}_camera_setting_controller"):
                getattr(
                    self, f"{microscope_name.lower()}_camera_setting_controller"
                ).update_experiment_values()
        if warning_message:
            return warning_message
        return ""

    def enable_resize(self):
        """Enable window resizing.
        
        """
        self.resize_ready_flag = True

    def resize(self, event):
        """Resize the GUI.

        Parameters
        __________
        event : event
            event = <Configure x=0 y=0 width=1200 height=600>
        """

        def refresh(width, height):
            """Refresh the GUI.

            Parameters
            __________
            width : int
                Width of the GUI.
            height : int
                Height of the GUI.
            """         
            self.view.scroll_frame.resize(width, height)
            self.view.right_frame.config(
                width=width-self.view.left_frame.winfo_width()-3,
                height=height-self.view.left_frame.winfo_height()
            )
            

        if not self.resize_ready_flag:
            return
        if event.widget != self.root:
            return
        if event.width == self.window_width and event.height == self.window_height:
            return

        if self.resize_event_id:
            self.view.after_cancel(self.resize_event_id)
        self.window_width = event.width
        self.window_height = event.height
        self.resize_event_id = self.view.after(
            300, lambda: refresh(event.width, event.height)
        )

    def prepare_acquire_data(self):
        """Prepare the acquisition data.

        Updates model.experiment.
        Sets sub-controller's mode to 'live' when 'continuous is selected, or 'stop'.

        Returns
        -------
        bool
            True if all settings are valid, False otherwise.
        """
        warning_info = self.update_experiment_setting()
        if warning_info:
            messagebox.showerror(
                title="Warning",
                message=f"Cannot start acquisition!\n{warning_info}",
            )
            return False

        # update real image width and height
        self.set_mode_of_sub(self.acquire_bar_controller.mode)
        self.update_buffer()
        return True

    def set_mode_of_sub(self, mode):
        """Communicates imaging mode to sub-controllers.

        Parameters
        __________
        mode : str
            The string = 'live', 'stop'
        """
        self.channels_tab_controller.set_mode(mode)
        self.camera_view_controller.set_mode(mode)
        self.camera_setting_controller.set_mode(mode)
        self.mip_setting_controller.set_mode(mode)
        self.waveform_tab_controller.set_mode(mode)

        # additional microscopes
        for microscope_name in self.additional_microscopes_configs:
            if hasattr(self, f"{microscope_name.lower()}_camera_setting_controller"):
                getattr(
                    self, f"{microscope_name.lower()}_camera_setting_controller"
                ).set_mode(mode)

        if mode == "stop":
            # GUI Failsafe
            self.acquire_bar_controller.stop_acquire()
            # self.menu_controller.feature_id_val.set(0)

    def execute(self, command, *args):
        """Functions listens to the Sub_Gui_Controllers.

        The controller.experiment is passed as an argument to the model, which then
        overwrites the model.experiment. Workaround due to model being in a sub-process.

        Parameters
        __________
        command : string
            The string includes 'stage', 'stop_stage', 'move_stage_and_update_info',
        args* : Iterable.
            Function-specific passes
        """

        if command == "stage":
            """Creates a thread and uses it to call the model to move stage

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            self.threads_pool.createThread(
                resourceName="model",
                target=self.move_stage,
                args=({args[1] + "_abs": args[0]},)
            )

        elif command == "stop_stage":
            """Creates a thread and uses it to call the model to stop stage"""
            self.threads_pool.createThread(
                resourceName="stop_stage",
                target=self.stop_stage)

        elif command == "query_stages":
            """Query the stages for their current position in a thread-blocking format.
            """
            query_thread = self.threads_pool.createThread(
                resourceName="model",
                target=self.stop_stage)

            while query_thread.is_alive():
                time.sleep(0.01)

        elif command == "move_stage_and_update_info":
            """update stage view to show the position

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            self.stage_controller.set_position(args[0])

        elif command == "move_stage_and_acquire_image":
            """update stage and acquire an image

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            stage_pos = dict(map(lambda axis: (axis + "_abs", args[0][axis]), args[0]))
            self.move_stage(stage_pos)
            self.update_stage_controller_silent(stage_pos)
            self.acquire_bar_controller.set_mode("single")
            self.execute("acquire")

        elif command == "get_stage_position":
            """Returns the current stage position from the widgets. 
            
            Does not communicate with the stages, but rather takes the last known 
            position.

            Returns
            -------
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            return self.stage_controller.get_position()

        elif command == "mark_position":
            """Appends a position to the multi-position list.

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
                values are in float64
            """
            self.multiposition_tab_controller.append_position(args[0])

        elif command == "resolution":
            """Changes the resolution mode and zoom position.

            Recalculates FOV_X and FOV_Y
            If Waveform Popup is open, communicates changes to it.

            Parameters
            ----------
            args : str
                "microscope_name zoom_value", "microscope_name", or "zoom_value"
            """
            # get microscope name and zoom value from args[0]
            temp = args[0].split()
            if len(temp) == 1:
                # microscope name is given
                if temp[0] in self.configuration_controller.microscope_list:
                    temp.append(
                        self.configuration_controller.get_zoom_value_list(temp[0])[0]
                    )
                elif temp[0] in self.configuration_controller.get_zoom_value_list(
                    self.configuration_controller.microscope_name
                ):
                    temp = [self.configuration_controller.microscope_name, temp[0]]
                else:
                    return
            resolution_value = " ".join(temp)
            if resolution_value != self.menu_controller.resolution_value.get():
                self.menu_controller.resolution_value.set(resolution_value)
                return
            self.change_microscope(temp[0], temp[1])
            work_thread = self.threads_pool.createThread(
                resourceName="model",
                target=lambda: self.model.run_command("update_setting", "resolution"))
            work_thread.join()

        elif command == "set_save":
            """Set whether the image will be saved.

            Parameters
            __________
            args : Boolean
                is_save = True/False
            """
            self.acquire_bar_controller.set_save_option(args[0])

        elif command == "update_setting":
            """Called by the Waveform Constants Popup Controller
            to update the Waveform constants settings in memory.

            Parameters
            __________
            args[0] : string
                string = 'resolution' or 'waveform' or 'galvo'...
            args[1] : dict
                dict = {
                'resolution_mode': self.resolution,
                'zoom': self.mag,
                'laser_info': self.resolution_info[
                'remote_focus_constants'][self.resolution][self.mag]
                }
            """
            self.threads_pool.createThread(
                resourceName="model",
                target=lambda: self.model.run_command("update_setting", *args)
            )

        elif command == "stage_limits":
            self.stage_controller.stage_limits = args[0]
            self.threads_pool.createThread(
                resourceName="model",
                target=lambda: self.model.run_command("stage_limits", *args)
            )

        elif command == "autofocus":
            """Execute autofocus routine."""
            if not self.acquire_bar_controller.is_acquiring:
                self.threads_pool.createThread(
                    resourceName="camera",
                    target=self.capture_image,
                    args=("autofocus", "live", *args),
                )
            elif self.acquire_bar_controller.mode == "live":
                self.threads_pool.createThread(
                    resourceName="model",
                    target=lambda: self.model.run_command("autofocus", *args)
                )

        elif command == "eliminate_tiles":
            """Execute eliminate tiles routine."""

            self.acquire_bar_controller.set_mode(mode="customized")
            feature_list = self.menu_controller.feature_list_names
            feature_name = "Remove Empty Tiles"
            try:
                # feature_id_val has a trace, and setting the menu item triggers it.
                feature_id = feature_list.index(feature_name)
                self.menu_controller.feature_id_val.set(feature_id)
            except ValueError:
                logger.debug("No feature named 'Remove Empty Tiles' found.")
                messagebox.showwarning(
                    title="Navigate", message="Feature 'Remove Empty Tiles' not found."
                )
                return
            self.execute("acquire")

        elif command == "load_feature":
            """Tell model to load/unload features."""

            work_thread = self.threads_pool.createThread(
                resourceName="model",
                target=lambda: self.model.run_command("load_feature", *args)
            )
            work_thread.join()

        elif command == "acquire_and_save":
            """Acquire data and save it.

            Prepares the acquisition data.
            Creates the file directory for saving the data.
            Saves the experiment file to that directory.
            Acquires the data.

            Parameters
            __________
            args[0] : dict
                dict = self.save_settings from the experiment.yaml file.

            """
            if not self.prepare_acquire_data():
                self.acquire_bar_controller.stop_acquire()
                return
            # get saving file directory
            file_directory = args[0]

            # Save the experiment.yaml file.
            save_yaml_file(
                file_directory=file_directory,
                content_dict=self.configuration["experiment"],
                filename="experiment.yml",
            )

            # Save the waveform_constants.yaml file.
            save_yaml_file(
                file_directory=file_directory,
                content_dict=self.configuration["waveform_constants"],
                filename="waveform_constants.yml",
            )

            # Save multi_positions.yml file
            save_yaml_file(
                file_directory=file_directory,
                content_dict=self.configuration["multi_positions"],
                filename="multi_positions.yml",
            )

            self.camera_setting_controller.solvent = self.configuration["experiment"][
                "Saving"
            ]["solvent"]
            self.camera_setting_controller.calculate_physical_dimensions()
            self.execute("acquire")

        elif command == "acquire":
            """Acquire data.

            Triggered when the Acquire button is hit by the user in the GUI.

            Prepares the acquisition data.

            Parameters
            __________
            args[0] : string
                string = 'continuous', 'z-stack', 'single', or 'projection'
            """
            # acquisition mode from plugin
            plugin_obj = self.plugin_acquisition_modes.get(
                self.acquire_bar_controller.mode, None
            )

            if plugin_obj and hasattr(plugin_obj, "prepare_acquisition_controller"):
                getattr(plugin_obj, "prepare_acquisition_controller")(self)

            # Prepare data
            if not self.prepare_acquire_data():
                self.acquire_bar_controller.stop_acquire()
                return

            # set the display segmentation flag to False
            self.camera_view_controller.display_mask_flag = False

            # ask user to verify feature list parameters if in "customized" mode
            if self.acquire_bar_controller.mode == "customized":
                feature_id = self.menu_controller.feature_id_val.get()
                if feature_id > 0:
                    if hasattr(self, "features_popup_controller"):
                        self.features_popup_controller.exit_func()
                    feature_list_popup = FeatureListPopup(
                        self.view, title="Feature List Configuration"
                    )
                    self.features_popup_controller = FeaturePopupController(
                        feature_list_popup, self
                    )
                    self.features_popup_controller.populate_feature_list(feature_id)

                    # wait until close the popup windows
                    self.view.wait_window(feature_list_popup.popup)

                    # do not run acquisition if "cancel" is selected
                    temp = self.features_popup_controller.start_acquisiton_flag
                    delattr(self, "features_popup_controller")
                    if not temp:
                        self.set_mode_of_sub("stop")
                        return

                    # if select 'ilastik segmentation' and 'show segmentation',
                    # TODO: update id if the feature id is changed
                    self.camera_view_controller.display_mask_flag = (
                        self.menu_controller.feature_id_val.get() == 4
                        and self.ilastik_controller.show_segmentation_flag
                    )

            self.stop_acquisition_flag = False
            self.launch_additional_microscopes()

            self.threads_pool.createThread(
                "camera",
                self.capture_image,
                args=(
                    "acquire",
                    self.acquire_bar_controller.mode,
                ),
            )

        elif command == "stop_acquire":
            """Stop the acquisition."""
            self.stop_acquisition_flag = True

            # self.model.run_command('stop')
            self.sloppy_stop()
            self.menu_controller.feature_id_val.set(0)

            # clear show_img_pipe
            while self.show_img_pipe.poll():
                self.show_img_pipe.recv()

            self.current_image_id = -1

        elif command == "exit":
            """Exit the program.

            Saves the current settings to .navigate/config/*.yml files.
            """
            self.sloppy_stop()
            self.update_experiment_setting()
            file_directory = os.path.join(get_navigate_path(), "config")
            for config_name, filename in [("experiment", "experiment.yml"),
                                          ("multi_positions", "multi_positions.yml"),
                                          ("gui", "gui_configuration.yml"),
                                          ("waveform_constants", "waveform_constants.yml"),
                                          ("rest_api_config", "rest_api_config.yml"),
                                          ("waveform_templates", "waveform_templates.yml")]:
                save_yaml_file(
                    file_directory=file_directory,
                    content_dict=self.configuration[config_name],
                    filename=filename,
                )

            self.model.run_command("terminate")
            self.model = None
            self.event_queue.put(("stop", ""))
            self.threads_pool.clear()
            sys.exit()

        # mirror commands:
        elif command in [
            "flatten_mirror",
            "zero_mirror",
            "set_mirror",
            "set_mirror_from_wcs",
        ]:
            self.threads_pool.createThread(
                "model", lambda: self.model.run_command(command, *args)
            )
        elif command == "tony_wilson":
            self.threads_pool.createThread(
                "camera",
                self.capture_image,
                args=(
                    "tony_wilson",
                    "live",
                ),
            )
        else:
            self.threads_pool.createThread(
                "model", lambda: self.model.run_command(command, *args)
            )

        # elif command == "change_camera":
        #     self.model.run_command("change_camera", *args)

        logger.info(
            f"Navigate Controller - command passed from child, {command}, {args}"
        )

    def sloppy_stop(self):
        """Keep trying to stop the model until successful.

        TODO: Delete this function!!!

        This is set up to get around the conflict between
        self.threads_pool.createThread('model', target)
        commands and the need to stop as abruptly as
        possible when the user hits stop. Here we leverage
        ObjectInSubprocess' refusal to let us access
        the model from two threads to our advantage, and just
        try repeatedly until we get a command in front
        of the next command in the model threads_pool resource.
        We should instead pause the model thread pool
        and interject our stop command, or clear the queue
        in threads_pool.
        """
        e = RuntimeError
        while e == RuntimeError:
            try:
                self.model.run_command("stop")
                e = None
            except RuntimeError:
                e = RuntimeError

    def capture_image(self, command, mode, *args):
        """Trigger the model to capture images.

        Parameters
        ----------
        command : string
            'acquire' or 'autofocus'
        mode : string
            'continuous', 'z-stack', 'single', or 'projection'
        args : function-specific passes.
        """
        self.camera_view_controller.image_count = 0
        self.mip_setting_controller.image_count = 0

        # Start up Progress Bars
        images_received = 0
        self.acquire_bar_controller.progress_bar(
            images_received=images_received,
            microscope_state=self.configuration["experiment"]["MicroscopeState"],
            mode=mode,
            stop=False,
        )
        try:
            work_thread = self.threads_pool.createThread(
                "model", lambda: self.model.run_command(command, *args)
            )
            work_thread.join()
        except Exception as e:
            messagebox.showerror(
                title="Error:",
                message=f"WARNING:\n{e}",
            )
            self.set_mode_of_sub("stop")
            return
        self.acquire_bar_controller.view.acquire_btn.configure(text="Stop")
        self.acquire_bar_controller.view.acquire_btn.configure(state="normal")
        microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]

        self.camera_view_controller.initialize_non_live_display(
            self.configuration["experiment"]["MicroscopeState"],
            self.configuration["experiment"]["CameraParameters"][microscope_name],
        )

        self.mip_setting_controller.initialize_non_live_display(
            self.configuration["experiment"]["MicroscopeState"],
            self.configuration["experiment"]["CameraParameters"][microscope_name],
        )

        self.stop_acquisition_flag = False
        start_time = time.time()
        self.camera_setting_controller.update_readout_time()

        while True:
            if self.stop_acquisition_flag:
                break
            # Receive the Image and log it.
            image_id = self.show_img_pipe.recv()
            logger.info(f"Navigate Controller - Received Image: {image_id}")

            if image_id == "stop":
                self.current_image_id = -1
                break

            self.current_image_id = image_id

            if not isinstance(image_id, int):
                logger.debug(
                    f"Navigate Controller - Something wrong happened, stop the model!, "
                    f"{image_id}"
                )
                self.execute("stop_acquire")

            # Display the image and update the histogram
            self.camera_view_controller.try_to_display_image(
                image=self.data_buffer[image_id]
            )
            self.mip_setting_controller.try_to_display_image(
                image=self.data_buffer[image_id]
            )
            self.histogram_controller.populate_histogram(
                image=self.data_buffer[image_id]
            )
            images_received += 1

            # Update progress bar.
            self.acquire_bar_controller.progress_bar(
                images_received=images_received,
                microscope_state=self.configuration["experiment"]["MicroscopeState"],
                mode=mode,
                stop=False,
            )
            # update framerate
            stop_time = time.time()
            try:
                frames_per_second = images_received / (stop_time - start_time)
            except ZeroDivisionError:
                frames_per_second = 1 / (
                    self.configuration["experiment"]["MicroscopeState"]["channels"][
                        "channel_1"
                    ].get("camera_exposure_time", 200)
                    / 1000
                )

            # Update the Framerate in the Camera Settings Tab
            self.camera_setting_controller.framerate_widgets["max_framerate"].set(
                frames_per_second
            )

            # Update the Framerate in the Acquire Bar to provide an estimate of
            # the duration of time remaining.
            self.acquire_bar_controller.framerate = frames_per_second

        logger.info(
            f"Navigate Controller - Captured {images_received}, " f"{mode} Images"
        )

        # acquisition mode from plugin
        plugin_obj = self.plugin_acquisition_modes.get(mode, None)
        if plugin_obj and hasattr(plugin_obj, "end_acquisition_controller"):
            getattr(plugin_obj, "end_acquisition_controller")(self)

        # Stop Progress Bars
        self.acquire_bar_controller.progress_bar(
            images_received=images_received,
            microscope_state=self.configuration["experiment"]["MicroscopeState"],
            mode=mode,
            stop=True,
        )
        self.set_mode_of_sub("stop")

    def launch_additional_microscopes(self):
        """Launch additional microscopes."""

        def display_images(
            microscope_name, camera_view_controller, show_img_pipe, data_buffer
        ):
            """Display images from additional microscopes.

            Parameters
            ----------
            microscope_name : str
                Microscope name
            camera_view_controller : CameraViewController
                Camera View Controller object.
            show_img_pipe : multiprocessing.Pipe
                The pipe for showing images.
            data_buffer : SharedNDArray
                Pre-allocated shared memory array.
                Size dictated by x_pixels, y_pixels, and number_of_frames in
                configuration file.
            """
            camera_view_controller.initialize_non_live_display(
                self.configuration["experiment"]["MicroscopeState"],
                self.configuration["experiment"]["CameraParameters"][microscope_name],
            )
            images_received = 0
            while True:
                if self.stop_acquisition_flag:
                    break
                # Receive the Image and log it.
                image_id = show_img_pipe.recv()
                logger.info(f"Navigate Controller - Received Image: {image_id}")

                if image_id == "stop":
                    break
                if not isinstance(image_id, int):
                    logger.debug(
                        f"Navigate Controller - Something wrong happened in additional "
                        f"microscope!, {image_id}"
                    )
                    break

                # Display the Image in the View
                try:
                    camera_view_controller.try_to_display_image(
                        image=data_buffer[image_id],
                    )
                except tkinter._tkinter.TclError:
                    print("Can't show images for the additional microscope!")
                    break
                images_received += 1

        # destroy all additional microscopes
        for microscope_name in list(self.additional_microscopes.keys()):
            destroy_window = False
            if microscope_name not in self.additional_microscopes_configs:
                destroy_window = True
            self.destroy_virtual_microscope(microscope_name, destroy_window)

        # show additional camera view popup
        for microscope_name in self.additional_microscopes_configs:
            show_img_pipe = self.model.create_pipe(f"{microscope_name}_show_img_pipe")
            data_buffer = self.model.launch_virtual_microscope(
                microscope_name,
                self.additional_microscopes_configs[microscope_name],
            )

            if microscope_name not in self.additional_microscopes:
                self.additional_microscopes[microscope_name] = {}

            if (
                "camera_view_controller"
                not in self.additional_microscopes[microscope_name]
            ):
                popup_window = CameraViewPopupWindow(self.view, microscope_name)
                camera_view_controller = CameraViewController(
                    popup_window.camera_view, self
                )
                camera_view_controller.microscope_name = microscope_name
                popup_window.popup.bind("<Configure>", camera_view_controller.resize)
                self.additional_microscopes[microscope_name][
                    "popup_window"
                ] = popup_window
                self.additional_microscopes[microscope_name][
                    "camera_view_controller"
                ] = camera_view_controller
                popup_window.popup.protocol(
                    "WM_DELETE_WINDOW",
                    combine_funcs(
                        popup_window.popup.dismiss,
                        lambda: self.additional_microscopes[microscope_name].pop(
                            "camera_view_controller"
                        ),
                    ),
                )

            self.additional_microscopes[microscope_name][
                "show_img_pipe"
            ] = show_img_pipe
            self.additional_microscopes[microscope_name]["data_buffer"] = data_buffer

            # start thread
            capture_img_thread = threading.Thread(
                target=display_images,
                args=(
                    microscope_name,
                    self.additional_microscopes[microscope_name][
                        "camera_view_controller"
                    ],
                    show_img_pipe,
                    self.additional_microscopes[microscope_name]["data_buffer"],
                ),
            )
            capture_img_thread.start()

    def destroy_virtual_microscope(self, microscope_name, destroy_window=True):
        """Destroy virtual microscopes.

        Parameters
        ----------
        microscope_name : str
            The microscope name
        destroy_window : bool
            The flag to dismiss window.
        """
        if microscope_name not in self.additional_microscopes:
            return
        del self.additional_microscopes[microscope_name]["data_buffer"]
        self.model.destroy_virtual_microscope(microscope_name)
        # release pipe
        self.model.release_pipe(f"{microscope_name}_show_img_pipe")
        del self.additional_microscopes[microscope_name]["show_img_pipe"]
        # destroy the popup window
        if destroy_window:
            self.additional_microscopes[microscope_name]["popup_window"].popup.dismiss()
            self.additional_microscopes[microscope_name][
                "camera_view_controller"
            ] = None
            del self.additional_microscopes[microscope_name]

    def move_stage(self, pos_dict):
        """Trigger the model to move the stage.

        Parameters
        ----------
        pos_dict : dict
            Dictionary of axis positions
        """
        # Update our local stage dictionary
        update_stage_dict(self, pos_dict)

        # Pass to model
        self.model.move_stage(pos_dict)

    def stop_stage(self):
        """Stop the stage.

        Grab the stopped position from the stage
        and update the GUI control values accordingly.
        """
        self.model.stop_stage()

    def update_stage_controller_silent(self, ret_pos_dict):
        """Send updates to the stage GUI

        Parameters
        ----------
        ret_pos_dict : dict
            Dictionary of axis positions
        """
        stage_gui_dict = {}
        for axis, val in ret_pos_dict.items():
            ax = axis.split("_")[0]
            stage_gui_dict[ax] = val
        self.stage_controller.set_position_silent(stage_gui_dict)

    def update_event(self):
        """Update the View/Controller based on events from the Model."""
        while True:
            event, value = self.event_queue.get()

            if event == "warning":
                # Display a warning that arises from the model as a top-level GUI popup
                messagebox.showwarning(title="Navigate", message=value)

            elif event == "multiposition":
                # Update the multi-position tab without appending to the list
                update_table(
                    table=self.multiposition_tab_controller.table,
                    pos=value,
                )
                self.channels_tab_controller.is_multiposition_val.set(True)

            elif event == "stop":
                # Stop the software
                break

            elif event == "update_stage":
                for _ in range(10):
                    try:
                        self.update_stage_controller_silent(value)
                        break
                    except RuntimeError:
                        time.sleep(0.001)
                        pass

            elif event in self.event_listeners.keys():
                try:
                    self.event_listeners[event](value)
                except Exception:
                    print(f"*** unhandled event: {event}, {value}")

    def add_acquisition_mode(self, name, acquisition_obj):
        """Add and Acquisition Mode.

        Parameters
        ----------
        name : string
            Name of the acquisition mode.
        acquisition_obj : object
            Object of the acquisition mode.
        """
        if name in self.plugin_acquisition_modes:
            print(f"*** plugin acquisition mode {name} exists, can't add another one!")
            return
        self.plugin_acquisition_modes[name] = acquisition_obj(name)
        self.acquire_bar_controller.add_mode(name)

    def register_event_listener(self, event_name, event_handler):
        """Register an event listener.

        Parameters
        ----------
        event_name : string
            Name of the event.
        event_handler : callable
            The function to handle the event.
        """
        self.event_listeners[event_name] = event_handler

    def register_event_listeners(self, events):
        """Register multiple event listeners.

        Parameters
        ----------
        events : dict
            Dictionary of event names and handlers.
        """
        for event_name, event_handler in events.items():
            self.register_event_listener(event_name, event_handler)



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Imports
import logging
from threading import Lock
from typing import Dict, Any

# Third Party Imports
from multiprocessing.managers import DictProxy, ListProxy

# Local Imports
from navigate.model.devices.daq.base import DAQBase
from navigate.model.devices.device_types import SerialDevice
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController
from navigate.tools.decorators import log_initialization


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class ASIDaq(DAQBase, SerialDevice):
    """ASIDAQ class for Data Acquisition (DAQ).

    Representation of Tiger Controller in action.
    Triggers all devices and outputs to camera trigger channel.
    """

    def __init__(
        self,
        microscope_name,
        device_connection,
        configuration: Dict[str, Any],
        device_id,
    ) -> None:
        """Initialize the ASI DAQ.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            Connection to the Tiger Controller.
        configuration : Dict[str, Any]
            Dictionary of configuration parameters.
        device_id : int
            Parameter necessary for start_device but not used here.
        """
        super().__init__(configuration)

        #: Lock: Lock for waiting to run.
        self.wait_to_run_lock = Lock()

        #: bool: Flag for updating analog task.
        self.is_updating_analog_task = False

        #: dict: Analog output tasks.
        self.analog_outputs = {}

        #: Any: Device connection.
        self.daq = device_connection

        #: str: microscope name
        self.microscope_name = microscope_name

        #: str: zoom
        self.zoom = self.configuration["experiment"]["MicroscopeState"]["zoom"]

        # retrieves galvo ListProxy/DictProxy from config file
        galvos_raw = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["galvo"]

        # Normalize galvos to always be a list of dicts
        if isinstance(galvos_raw, DictProxy):
            self.galvos = [dict(galvos_raw)]  # wrap single DictProxy into a list
        elif isinstance(galvos_raw, ListProxy):
            self.galvos = [
                dict(g) for g in galvos_raw
            ]  # convert each DictProxy in the list to a dict
        else:
            raise TypeError("Unexpected type for galvos: {}".format(type(galvos_raw)))

        # update analog outputs with galvo IDs and associated axes
        i = 0
        for g in self.galvos:
            self.analog_outputs[f"galvo {i}"] = g["hardware"]["axis"]
            i += 1

        # retreive galvo phases from config
        self.phases = [galvo["phase"] for galvo in self.galvos]

        # update analog outputs with rfvc and associated axis
        remote_focus_channel = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["remote_focus"]["hardware"]["axis"]
        self.analog_outputs["remote_focus"] = remote_focus_channel

        # sets up initial PLC configuration with default delay (ms), camera delay, rfvc delay, sweep time (ms), and analog outputs dict
        self.daq.setup_control_loop([200], 0, 0, 120, self.analog_outputs)

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASIDaq Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, default is 115200.
        timeout : float
            Timeout for communicating with the filter wheel, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI stage connection failed.")
            raise Exception("ASI stage connection failed.")
        return tiger_controller

    def prepare_acquisition(self, channel_key: str) -> None:
        """Prepare the acquisition.

        Creates and configures the DAQ tasks.
        Writes the waveforms to each task.

        Parameters
        ----------
        channel_key : str
            Channel key for current channel.
        """
        # Get appropriate sweep_time for the current channel
        sweep_time = self.sweep_times[channel_key]

        # loop through galvo phases to calculate time delays
        # delays[i] corresponds to the time between the ith galvo trigger and the master trigger
        n = len(self.galvos)
        i = 0
        delays = []
        for phase in self.phases:
            frequency = self.waveform_constants["galvo_constants"][f"Galvo {i}"][
                self.microscope_name
            ][self.zoom]["frequency"]
            period = self.exposure_times[channel_key] * 1000 / float(frequency)

            # round period for triangle waveform to even number, as the TG-1000 can only generate triangle waveforms with even-number-of-ms periods
            if self.galvos[i]["waveform"] == "sawtooth":
                rising_ramp = float(
                    self.waveform_constants["galvo_constants"][f"Galvo {i}"][
                        self.microscope_name
                    ][self.zoom].get("rising_ramp", 50)
                )
                if rising_ramp == 50:
                    period = 2 * round(period / 2)

            # round period to closest number of ms, as TG-1000 can only generate waveforms with whole-number-of-ms periods
            period = int(round(period))

            # save first period to sync control loop with first galvo
            if i == 0:
                period0 = period

            # calculate time delay corresponding to phase offset
            t = period * phase / (2 * 3.14159265)
            delays.append(period * (n - i) - t)
            i += 1

        # modify sweep time to sync with the first galvo if there are asi galvos in config
        if len(delays) > 0:
            n7 = 1000 * sweep_time // period0 + 1
            sweep_time = period0 * n7

        self.camera_delay = float(
            self.waveform_constants["other_constants"].get("camera_delay", 5)
        )
        rfvc_delay = float(
            self.waveform_constants["other_constants"].get("remote_focus_delay", 5)
        )
        # sets up control loop with all parameters (all times in ms)
        self.daq.setup_control_loop(
            delays, self.camera_delay, rfvc_delay, sweep_time, self.analog_outputs
        )

        self.current_channel_key = channel_key
        self.is_updating_analog_task = False
        # if self.wait_to_run_lock.locked():
        #     self.wait_to_run_lock.release()

    def run_acquisition(self) -> None:
        """Run DAQ Acquisition.

        Run the tasks for triggering, analog and counter outputs.
        The master trigger initiates all other waveforms via a shared trigger
        For this to work, all analog output and counter tasks have to be primed so that
        they are waiting for the trigger signal.
        """
        # if self.is_updating_analog_task:
        #     self.wait_to_run_lock.acquire()
        #     self.wait_to_run_lock.release()

        # turn on PLC cell 1 (Master Trigger)
        try:
            self.daq.logic_cell_on("1")
        except Exception:
            logger.debug("DAQ cannot turn on")
            pass

    def stop_acquisition(self) -> None:
        """Stop Acquisition.

        Stop control loop.
        """
        # turn on PLC cell 8 (kills control loop)
        # reset PLC cell 1 (Master Trigger)
        try:
            self.daq.logic_cell_on("8")
            self.daq.logic_cell_off("1")
        except Exception:
            logger.debug("DAQ cannot turn off")
            pass

        # if self.wait_to_run_lock.locked():
        #     self.wait_to_run_lock.release()
